from __future__ import annotations

from collections import defaultdict
from logging import getLogger
from functools import total_ordering
from pathlib import Path
import re
from typing import Set, Iterable, Tuple, Dict
from urllib.parse import quote_plus

from IPython.display import display, Markdown
from lxml import etree
from tqdm import tqdm

from ..util import normalize_ocr
from ..config import CONFIG as _CONFIG


CONFIG = _CONFIG['entity.Regest']
LOGGER = getLogger(__name__)

COMPOSITE_ENTITY = re.compile(r'(?P<before_parens>.*)\((?P<inside_parens>[^)]*)\)\s*')


@total_ordering
class Regest:
    ROOT_PATH = Path(CONFIG['root_path'])
    DOCUMENT_URL = CONFIG['document_url']

    def __init__(self, filename: str):
        self.filename = filename
        self.basename = str(Path(filename).with_suffix(''))
        self.document_url = self.DOCUMENT_URL.format(basename=quote_plus(self.basename))
        self.text_filename = self.ROOT_PATH / Path(self.filename).with_suffix('.xml')
        entities: Set['Entity'] = set()
        parser = etree.XMLParser(huge_tree=True, encoding='utf-8')
        with self.text_filename.open('rb') as f:
            document = etree.parse(f, parser)
        for entity in document.xpath('//placeName'):
            if entity.attrib['reg']:
                entities |= set(Place.from_text(entity.attrib['reg']))
        for entity in document.xpath('//persName'):
            if entity.attrib['reg']:
                entities |= set(Person.from_text(entity.attrib['reg']))
        self.entities = entities

    def __hash__(self):
        return hash(self.basename)

    def __eq__(self, other) -> bool:
        if isinstance(other, Regest):
            return self.basename == other.basename
        return NotImplemented

    def __lt__(self, other) -> bool:
        if isinstance(other, Regest):
            return self.basename < other.basename
        return NotImplemented

    def __str__(self) -> str:
        return self.basename

    def __repr__(self) -> str:
        return str(self)

    def _ipython_display_(self) -> None:
        summary = 'Regest `{}` with {} entities'.format(self.basename, len(self.entities))
        display(Markdown(summary))


@total_ordering
class Entity:
    def __init__(self, text: str):
        if not text:
            raise ValueError('Empty entity')
        self.text = normalize_ocr(text)

    @classmethod
    def from_text(entity_class: type['Entity'], text: str) -> Iterable['Entity']:
        match = re.fullmatch(COMPOSITE_ENTITY, text)
        if match is None:
            yield entity_class(text)
        else:
            yield entity_class(match.group('before_parens'))
            yield entity_class(match.group('inside_parens'))

    def __len__(self) -> int:
        return len(str(self))

    def __hash__(self):
        return hash((self.__class__.__name__, str(self)))

    def __eq__(self, other) -> bool:
        if isinstance(other, Entity):
            return repr(self) == repr(other)
        return NotImplemented

    def __lt__(self, other) -> bool:
        if isinstance(other, Entity):
            return repr(self) < repr(other)
        return NotImplemented

    def __str__(self) -> str:
        return self.text

    def __repr__(self) -> str:
        return '{}: {}'.format(self.__class__.__name__, self)

    @staticmethod
    def from_repr(entity_repr: str) -> 'Entity':
        if entity_repr.startswith('Place: '):
            return Place(entity_repr[7:])
        else:
            assert entity_repr.startswith('Person: ')
            return Person(entity_repr[8:])

    def _ipython_display_(self) -> None:
        display(Markdown(str(self)))


class Place(Entity):
    def __init__(self, text):
        super().__init__(text)


class Person(Entity):
    def __init__(self, text):
        super().__init__(text)


def entity_factory(filename: str) -> Tuple[Regest, Set[Entity]]:
    filename = str(Path(filename).relative_to(Regest.ROOT_PATH))
    regest = Regest(filename)
    return (regest, regest.entities)


def load_entities() -> Dict[Entity, Set[Regest]]:
    filenames = sorted(Regest.ROOT_PATH.glob('*.xml'))
    places, persons = defaultdict(lambda: set()), defaultdict(lambda: set())
    all_entities = defaultdict(lambda: set())
    for regest, entities in map(entity_factory, tqdm(filenames, desc='Loading entities')):
        for entity in entities:
            if isinstance(entity, Place):
                places[entity].add(regest)
                all_entities[entity].add(regest)
            else:
                assert isinstance(entity, Person)
                persons[entity].add(regest)
                all_entities[entity].add(regest)
    assert len(all_entities) == len(places) + len(persons)
    message = 'Loaded {} entities: {} places ({:.2f}%) and {} persons ({:.2f}%).'
    message = message.format(
        len(all_entities),
        len(places),
        100.0 * len(places) / len(all_entities),
        len(persons),
        100.0 * len(persons) / len(all_entities),
    )
    LOGGER.info(message)
    return all_entities
